#include <iostream>
#include <string>
#include <sstream>
#include <fstream>

#include "Compiler.h"
#include "Script.h"
#include "ScriptData.h"

using namespace std;

Compiler::Compiler(istream& input): input(input), code(NULL), codeP(NULL), codeE(NULL)
{
}

Compiler::~Compiler()
{
	delete [] code;
}

bool Compiler::Compile()
{
	while(input.good())
	{
		CompileStatement();
		++line;
	}
	
	return errStream.str().length() == 0;
}

void Compiler::CompileStatement()
{
	string command;
	input >> command;
	if(command.length() == 0)
		return;
	if(command == "set")
	{
		string target;
		input >> target;
		if(target == "context")
		{
			int rtarget;
			input >> target;
			if(target == "object")
				rtarget = 0;
			else if(target == "player")
				rtarget = 1;
			else
			{
				errStream << "Unknown context: " << target << endl;
				EatLine();
				return;
			}
			EmitOpcode(Opcodes::SetContext);
			EmitOpcode(Opcodes::IntegerConstant);
			EmitInteger(rtarget);
		}
		else if(target == "game")
		{
			string var;
			int val;
			input >> var >> val;
			if(!input.good())
			{
				errStream << "Unexpected end of file" << endl;
				return;
			}
			EmitOpcode(Opcodes::SetGameVariable);
			EmitOpcode(Opcodes::StringConstant);
			EmitString(var);
			EmitOpcode(Opcodes::IntegerConstant);
			EmitInteger(val);
		}
		else if(target == "scene")
		{
			string var;
			int val;
			input >> var >> val;
			if(!input.good())
			{
				errStream << "Unexpected end of file" << endl;
				return;
			}
			EmitOpcode(Opcodes::SetSceneVariable);
			EmitOpcode(Opcodes::StringConstant);
			EmitString(var);
			EmitOpcode(Opcodes::IntegerConstant);
			EmitInteger(val);
		}
		else
		{
			errStream << "Unknown target for set: " << target << endl;
			EatLine();
			return;
		}
	}
	else if(command == "if")
	{
		EmitOpcode(Opcodes::JumpZero);
		CompileExpression();
		EmitOpcode(Opcodes::IntegerConstant);
		int branch_loc = codeP - code;
		EmitInteger(0);
		while(true)
		{
			string next;
			NextWord(next);
			if(!input.good())
			{
				errStream << "Unexpected end of file" << endl;
				return;
			}
			if(next == "else")
			{
				EmitOpcode(Opcodes::Jump);
				EmitOpcode(Opcodes::IntegerConstant);
				int branch2_loc = codeP - code;
				EmitInteger(0);
				int loc = codeP - code;
				codeP = code + branch_loc;
				EmitInteger(loc);
				codeP = code + loc;
				branch_loc = branch2_loc;
				ws(input);
				EatLine();
			}
			if(next == "end")
			{
				int loc = codeP - code;
				codeP = code + branch_loc;
				EmitInteger(loc);
				codeP = code + loc;
				ws(input);
				EatLine();
				return;
			}
			CompileStatement();
		}
	}
	else if(command == "rest")
	{
		EmitOpcode(Opcodes::Rest);
		CompileInteger();
	}
	else if(command == "color")
	{
		EmitOpcode(Opcodes::Color);
		CompileInteger();
		CompileInteger();
		CompileInteger();
	}
	else if(command == "clear")
	{
		EmitOpcode(Opcodes::Clear);
	}
	else if(command == "clear_print")
	{
		EmitOpcode(Opcodes::ClearPrint);
	}
	else if(command == "print_center")
	{
		EmitOpcode(Opcodes::PrintCenter);
		CompileInteger();
		CompileString();
	}
	else if(command == "print")
	{
		EmitOpcode(Opcodes::Print);
		CompileString();
	}
	else if(command == "destroy")
	{
		EmitOpcode(Opcodes::Destroy);
	}
	else if(command == "hide")
	{
		EmitOpcode(Opcodes::Hide);
	}
	else if(command == "load")
	{
		EmitOpcode(Opcodes::Load);
		CompileString();
	}
	else
	{
		errStream << "Unknown command " << command << endl;
		EatLine();
	}
}

void Compiler::CompileInteger()
{
	int in;
	input >> in;
	EmitOpcode(Opcodes::IntegerConstant);
	EmitInteger(in);
}

void Compiler::CompileString()
{
	string str;
	ws(input);
	while(true)
	{
		int c = input.get();
		if(!input.good() || c == '\n')
			break;
		str += c;
	}
	EmitOpcode(Opcodes::StringConstant);
	EmitString(str);
}

void Compiler::CompileExpression()
{
	string scope;
	input >> scope;
	if(scope == "scene")
	{
		string var;
		input >> var;
		EmitOpcode(Opcodes::GetSceneVariable);
		EmitOpcode(Opcodes::StringConstant);
		EmitString(var);
	}
	else if(scope == "game")
	{
		string var;
		input >> var;
		EmitOpcode(Opcodes::GetGameVariable);
		EmitOpcode(Opcodes::StringConstant);
		EmitString(var);
	}
	else
	{
		errStream << "Unknown scope " << scope;
		EatLine();
	}
}

void Compiler::EmitOpcode(int index)
{
	//!! ENDIANESS!
	ReserveCode(2);
	*((short*) codeP) = index;
	codeP += 2;
}

void Compiler::EmitInteger(int in)
{
	ReserveCode(4);
	*codeP++ = (in >> 24) & 0xff;
	*codeP++ = (in >> 16) & 0xff;
	*codeP++ = (in >> 8) & 0xff;
	*codeP++ = in & 0xff;
}

void Compiler::EmitString(const string& str)
{
	ReserveCode(str.length() + 1);
	memcpy(codeP, str.c_str(), str.length() + 1);
	codeP += str.length() + 1;
}

void Compiler::ShrinkCode()
{
	int offs = codeP - code;
	unsigned char* newCode = new unsigned char [offs];
	memcpy(newCode, code, offs);
	delete [] code;
	code = newCode;
	codeP = code + offs;
	codeE = code + offs;
}

void Compiler::ReserveCode(int extra)
{
	if(code == NULL)
	{
		code = new unsigned char [2048];
		assert(code);
		codeP = code;
		codeE = code + 2048;
		return;
	}
	if(codeE - codeP <= extra)
	{
		int offs = codeP - code;
		int size = codeE - code;
		unsigned char* newCode = new unsigned char [2048 + size];
		memcpy(newCode, code, codeE - code);
		delete [] code;
		code = newCode;
		codeP = code + offs;
		codeE = code + size;
	}
}

void Compiler::EatLine()
{
	while(true)
	{
		int c = input.get();
		if(!input.good() || c == '\n')
			break;
	}
}

void Compiler::NextWord(string& next)
{
	int pos = input.tellg();
	input >> next;
	input.seekg(pos);
}
