#include <allegro.h>
#include <assert.h>

#include <algorithm>
#include <string>
#include <list>
#include <iostream>

#include "Scene.h"
#include "Game.h"
#include "Object.h"
#include "SceneryObject.h"

using namespace std;

static void saveMask(PACKFILE* pf, BITMAP* data)
{
	pack_iputw(data->w, pf);
	pack_iputw(data->h, pf);
	char* buffer = new char [(data->w * data->h + 7) / 8];
	memset(buffer, 0, (data->w * data->h + 7) / 8);
	int byte = 0, bit = 0;
	for(int y = 0; y < data->h; y++)
	{
		for(int x = 0; x < data->w; x++)
		{
			if(getpixel(data, x, y))
				buffer[byte] |= 1 << (7 - bit);
			bit++;
			if(bit > 7)
			{
				byte++;
				bit = 0;
			}
		}
	}
	pack_fwrite(buffer, (data->w * data->h + 7) / 8, pf);
	delete [] buffer;
}

static void loadMask(PACKFILE* pf, BITMAP** data)
{
	int w = pack_igetw(pf);
	int h = pack_igetw(pf);
	*data = create_bitmap(w, h);
	char* buffer = new char [((*data)->w * (*data)->h + 7) / 8];
	pack_fread(buffer, ((*data)->w * (*data)->h + 7) / 8, pf);
	int byte = 0, bit = 0;
	for(int y = 0; y < (*data)->h; y++)
	{
		for(int x = 0; x < (*data)->w; x++)
		{
			if(buffer[byte] & (1 << (7 - bit)))
				putpixel((*data), x, y, -1);
			bit++;
			if(bit > 7)
			{
				byte++;
				bit = 0;
			}
		}
	}
	delete [] buffer;
}

static bool SortObjects(Object *lhs, Object *rhs)
{
	return lhs->y + lhs->z < rhs->y + rhs->z;
}

Scene::Scene(Game* myGame) : background(NULL), collision(NULL), player(NULL), panX(0), panY(0), myGame(myGame)
{
}

Scene::~Scene()
{
	destroy_bitmap(background);
	destroy_bitmap(collision);
	
	for(vector<Entry*>::iterator i = entries.begin(); i != entries.end(); i++)
		delete *i;
	
	for(list<Object*>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		if((*i)->sceneOwned)
			delete *i;
	}
}


bool Scene::Load(const string& location)
{
	PACKFILE* pf = pack_fopen(location.c_str(), "rp");
	if(!pf)
	{
		cerr << "Failed to open " << location << endl;
		return false;
	}
	
	title = location;
	
	int length;
	char* buffer;
	
	length = pack_igetw(pf);
	for(int i = 0; i < length; i++)
	{
		Entry* e = new Entry();
		int length2;
		length2 = pack_igetw(pf);
		buffer = new char[length2 + 1];
		pack_fread(buffer, length2, pf);
		buffer[length2] = '\0';
		e->name = buffer;
		delete [] buffer;
		
		e->x = pack_igetw(pf);
		e->y = pack_igetw(pf);
		e->z = pack_igetw(pf);
		entries.push_back(e);
	}
	
	length = pack_igetw(pf);
	for(int i = 0; i < length; i++)
	{
		SceneryObject* o = new SceneryObject();
		
		int hasDisplay = pack_getc(pf);
		if(hasDisplay)
		{
			o->bmp = load_pcx_pf(pf, NULL);
		}
		else
			o->bmp = NULL;
			
		o->x = pack_igetw(pf);
		o->y = pack_igetw(pf);
		o->z = pack_igetw(pf);
		o->w = pack_igetw(pf);
		o->h = pack_igetw(pf);
		o->give = pack_getc(pf);
		o->obstacle = pack_getc(pf);
		o->sceneOwned = true;
		o->scene = this;
		
		int length2;
		length2 = pack_igetw(pf);
		for(int j = 0; j < length2; j++)
		{
			string name;
			Script data;
			
			int length3 = pack_igetw(pf);
			buffer = new char[length3 + 1];
			pack_fread(buffer, length3, pf);
			buffer[length3] = '\0';
			name = buffer;
			delete [] buffer;
		
			bool res = data.Load(o, pf);
			assert(res);
			data.SetScene(this);
			
			Event e(name, data);
			o->events.push_back(e);
		}
		
		objects.push_back(o);
	}
	
	background = load_pcx_pf(pf, NULL);
	loadMask(pf, &collision);
	
	length = pack_igetw(pf);
	for(int i = 0; i < length; i++)
	{
		string name;
		Script data;
		
		int length2 = pack_igetw(pf);
		buffer = new char[length2 + 1];
		pack_fread(buffer, length2, pf);
		buffer[length2] = '\0';
		name = buffer;
		delete [] buffer;
	
		bool res = data.Load(NULL, pf);
		data.SetScene(this);
		assert(res);
		
		Event e(name, data);
		events.push_back(e);
	}
	
	pack_fclose(pf);
	return true;
}

bool Scene::Save(const string& location)
{
#ifndef PACKFILE_COMPRESSION
	const char* mode = "w!";
#else
	const char* mode = "wp";
#endif
	PACKFILE* pf = pack_fopen(location.c_str(), mode);
	if(!pf)
	{
		cerr << "Failed to create " << location << endl;
		return false;
	}
	
	title = location;
	
	pack_iputw(entries.size(), pf);
	for(vector<Entry*>::iterator i = entries.begin(); i != entries.end(); i++)
	{
		pack_iputw((*i)->name.length(), pf);
		pack_fwrite((*i)->name.c_str(), (*i)->name.length(), pf);
		pack_iputw((*i)->x, pf);
		pack_iputw((*i)->y, pf);
		pack_iputw((*i)->z, pf);
	}
	
	pack_iputw(objects.size(), pf);
	for(list<Object*>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		// Only save scenery
		SceneryObject* so = dynamic_cast<SceneryObject*>(*i);
		if(!so)
			continue;
			
		if(so->bmp)
		{
			pack_putc(1, pf);
			save_pcx_pf(pf, so->bmp, NULL);
		}
		else
			pack_putc(0, pf);
		pack_iputw(so->x, pf);
		pack_iputw(so->y, pf);
		pack_iputw(so->z, pf);
		pack_iputw(so->w, pf);
		pack_iputw(so->h, pf);
		pack_putc(so->give, pf);
		pack_putc(so->obstacle, pf);
		
		pack_iputw(so->events.size(), pf);
		for(vector<Event>::iterator j = so->events.begin(); j != so->events.end(); j++)
		{
			pack_iputw(j->name.length(), pf);
			pack_fwrite(j->name.c_str(), j->name.length(), pf);
			j->data.Save(pf);
		}
	}
	
	save_pcx_pf(pf, background, NULL);
	saveMask(pf, collision);
	
	pack_iputw(events.size(), pf);
	for(vector<Event>::iterator i = events.begin(); i != events.end(); i++)
	{
		pack_iputw(i->name.length(), pf);
		pack_fwrite(i->name.c_str(), i->name.length(), pf);
		i->data.Save(pf);
	}
	
	pack_fclose(pf);
	return true;
}

void Scene::SetPlayer(Character* player)
{
	if(this->player)
		objects.remove(this->player);
	objects.push_back(player);
	this->player = player;
	SortObjects();
}

bool Scene::MoveToEntry(Object* object, const string& entry)
{
	for(vector<Entry*>::iterator i = entries.begin(); i != entries.end(); i++)
	{
		if((*i)->name == entry)
		{
			object->x = (*i)->x;
			object->y = (*i)->y;
			object->z = (*i)->z;
			SortObjects();
			return true;
		}
	}
	return false;
}

bool Scene::MoveObject(Object* object, int x, int y)
{
	if(getpixel(collision, x, y) != -1)
		return false;
	if(getpixel(collision, x + object->GetBitmap()->w, y) != -1)
		return false;
	for(list<Object*>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		if(object == *i || !(*i)->obstacle)
			continue;
		if((*i)->x <= x && (*i)->x + (*i)->w > x && (*i)->y - (*i)->h <= y && (*i)->y > y)
			return false;
	}
	object->x = x;
	object->y = y;
	SortObjects();
	return true;
}

void Scene::SortObjects()
{
	objects.sort(::SortObjects);
}

bool Scene::FireEvent(const string& name)
{
	for(vector<Event>::iterator i = events.begin(); i != events.end(); i++)
	{
		if(i->name == name)
		{
			i->data.Reset();
			i->running = true;
			return true;
		}
	}
	return false;
}

void Scene::TryInteract(int x, int y)
{
	for(list<Object*>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		if((*i)->x - (*i)->give <= x && (*i)->x + (*i)->w + (*i)->give > x && (*i)->y - (*i)->h - (*i)->give <= y && (*i)->y + (*i)->give > y)
		{
			(*i)->FireEvent("interact");
		}
	}
}

void Scene::GoToScene(const string& filename)
{
	myGame->GoToScene(filename);
}

void Scene::SetGameVariable(const string& var, int val)
{
	myGame->gameVars[var] = val;
}

void Scene::SetSceneVariable(const string& var, int val)
{
	myGame->sceneVars[title][var] = val;
}

int Scene::GetGameVariable(const string& var)
{
	return myGame->gameVars[var];
}

int Scene::GetSceneVariable(const string& var)
{
	return myGame->sceneVars[title][var];
}

void Scene::Logic()
{
	assert(player != NULL);
	for(list<Object*>::iterator i = objects.begin(); i != objects.end();)
	{
		(*i)->Logic();
		if((*i)->deleteMe)
			i = objects.erase(i);
		else
			i++;
	}
	
	for(std::vector<Event>::iterator i = events.begin(); i != events.end(); i++)
	{
		if(i->running)
			i->running = i->data.Logic();
	}
	
	int fx = player->x + player->GetBitmap()->w / 2, fy = player->y + player->GetBitmap()->h / 2;
	
	if(fx - panX > 320 / 6)
		panX = fx - 320 / 6;
	if(fx - panX < -320 / 6)
		panX = fx + 320 / 6;
	if(panX < 320 / 2)
		panX = 320 / 2;
	if(panX >= background->w - 320 / 2)
		panX = background->w - 320 / 2 - 1;
		
	if(fy - panY > 240 / 3)
		panY = fy - 240 / 3;
	if(fy - panY < -240 / 3)
		panY = fy + 240 / 3;
	if(panY < 240 / 2)
		panY = 240 / 2;
	if(panY >= background->h - 240 / 2)
		panY = background->h - 240 / 2 - 1;
}

void Scene::Draw(BITMAP *dest)
{
	int px = panX - dest->w / 2, py = panY - dest->h / 2;
		
	blit(background, dest, px, py, 0, 0, background->w, background->h);

	for(list<Object*>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		BITMAP *b = (*i)->GetBitmap();
		if(!b)
			continue;
			
		masked_blit(b, dest, 0, 0, (*i)->x - px, (*i)->y - py - b->h, b->w, b->h);
		
		if((*i)->message.length() != 0)
		{
			int length = text_length(font, (*i)->message.c_str());
			int sx = (*i)->x + b->w / 2 - length / 2 - px;
			int sy = (*i)->y - b->h - 10 - py;
			
			if(sx + length >= 320)
				sx = 320 - length - 1;
			if(sx < 0)
				sx = 0;
			if(sy >= 220)
				sy = 219;
			if(sy < 2)
				sy = 2;
			
			textprintf_ex(dest, font, sx, sy, (*i)->color, -1, "%s", (*i)->message.c_str());
		}
	}
	
	for(list<Object*>::iterator i = objects.begin(); i != objects.end(); i++)
		(*i)->DrawScripts(dest);
	
	for(vector<Event>::iterator i = events.begin(); i != events.end(); i++)
	{
		if(i->running && i->data.plane)
		{
			masked_blit(i->data.plane, dest, 0, 0, 0, 0, i->data.plane->w, i->data.plane->h);
		}
	}
}
