#include "ScriptData.h"
#include "Constants.h"
#include "Script.h"
#include "Object.h"
#include "Scene.h"
#include "Game.h"

#define DEFINE_OPCODES
#include "Opcodes.h"

#include <iostream>
using namespace std;

#define PARAMETER(name, type) WeakType name; Step(name); assert(name.GetType() == WeakType::type)

ScriptData::ScriptData(Script* script): code(NULL), codeP(NULL), codeE(NULL), script(script)
{
}

ScriptData::ScriptData(Script* script, const ScriptData& other): code(NULL), codeP(NULL), codeE(NULL), script(script)
{
	int size = other.codeE - other.code;
	code = new unsigned char [size];
	memcpy(code, other.code, size);
	codeE = code + (other.codeE - other.code);
	codeP = code + (other.codeP - other.code);
}

ScriptData::~ScriptData()
{
	delete [] code;
}

bool ScriptData::Step()
{
	if(codeP >= codeE)
		return false;
	WeakType result;
	Step(result);
	return true;
}

void ScriptData::Reset()
{
	codeP = code;
}

void ScriptData::Step(WeakType& result)
{
	assert(codeE - codeP >= 2);
	//!! ENDIANESS!
	short index = *(short*) codeP;
	codeP += 2;
	(this->*(opcodes[index]))(result);
}

void ScriptData::VoidConstant(WeakType& result)
{
	assert(false);
}

void ScriptData::IntegerConstant(WeakType& result)
{
	unsigned int res;
	assert(codeE - codeP >= 4);
	res = *codeP++ << 24;
	res |= *codeP++ << 16;
	res |= *codeP++ << 8;
	res |= *codeP++;
	result = WeakType(res);
}

void ScriptData::StringConstant(WeakType& result)
{
	string res;
	while(true)
	{
		assert(codeP != codeE);
		char c = *codeP++;
		if(c == '\0')
			break;
		res += c;
	}
	result = WeakType(res);
}

void ScriptData::SetContext(WeakType& result)
{
	PARAMETER(target, T_Integer);
	assert((int) target >= 0);
	assert((int) target < 2);
	switch((int) target)
	{
		case 0:
			script->context = script->saveContext;
			break;
		case 1:
			script->context = script->scene->myGame->player;
			break;
	}
}

void ScriptData::Jump(WeakType& result)
{
	PARAMETER(offs, T_Integer);
	codeP = code + (int) offs;
}

void ScriptData::JumpZero(WeakType& result)
{
	PARAMETER(test, T_Integer);
	PARAMETER(offs, T_Integer);
	if(test == 0)
		codeP = code + (int) offs;
}

void ScriptData::Rest(WeakType& result)
{
	PARAMETER(time, T_Integer);
	script->sleeping = time * LOGIC_RATE / 1000;
}

void ScriptData::Color(WeakType& result)
{
	PARAMETER(r, T_Integer);
	PARAMETER(g, T_Integer);
	PARAMETER(b, T_Integer);
	assert(script->context);
	script->context->color = makecol(r, g, b);
}

void ScriptData::Clear(WeakType& result)
{
	clear_to_color(script->plane, makecol(255, 0, 255));
}

void ScriptData::ClearPrint(WeakType& result)
{
	assert(script->context);
	script->context->message = "";
}

void ScriptData::PrintCenter(WeakType& result)
{
	PARAMETER(line, T_Integer);
	PARAMETER(string, T_String);
	int y = (int) line * 10 + script->plane->h / 2;
	textprintf_centre_ex(script->plane, font, script->plane->w / 2, y, -1, -1, "%s", (const char*) string);
}

void ScriptData::Print(WeakType& result)
{
	PARAMETER(string, T_String);
	assert(script->context);
	script->context->message = (const char*) string;
}

void ScriptData::SetGameVariable(WeakType& result)
{
	PARAMETER(var, T_String);
	PARAMETER(val, T_Integer);
	assert(script->scene);
	script->scene->SetGameVariable((const char*) var, (int) val);
}

void ScriptData::GetGameVariable(WeakType& result)
{
	PARAMETER(var, T_String);
	assert(script->scene);
	result = WeakType(script->scene->GetGameVariable((const char*) var));
}

void ScriptData::SetSceneVariable(WeakType& result)
{
	PARAMETER(var, T_String);
	PARAMETER(val, T_Integer);
	assert(script->scene);
	script->scene->SetSceneVariable((const char*) var, (int) val);
}

void ScriptData::GetSceneVariable(WeakType& result)
{
	PARAMETER(var, T_String);
	assert(script->scene);
	result = WeakType(script->scene->GetSceneVariable((const char*) var));
}

void ScriptData::Destroy(WeakType& result)
{
	assert(script->scene);
	script->context->deleteMe = true;
}

void ScriptData::Hide(WeakType& result)
{
	assert(script->scene);
	script->context->y = -1;
}

void ScriptData::Load(WeakType& result)
{
	PARAMETER(loc, T_String);
	assert(script->scene);
	script->scene->GoToScene((const char*) loc);
}
